# sb::Driver::TextOld
# == Author(s) : Takuya Otani <takuya.otani@gmail.com> ==
# == Copyright (C) 2005 SimpleBoxes/SerendipityNZ Ltd. ==

# ======================================
package sb::Driver::TextOld;
use strict;
use sb::Lock ();
use sb::File ();
use sb::Language ();
use sb::Driver ();
use base qw( sb::Driver );
use vars qw( $VERSION );
$VERSION = '0.16';
# 0.16 [2008/07/20] added alter
# 0.15 [2007/12/04] added insert / modified setup to create a lock file
# 0.14 [2007/08/12] modified error strings
# 0.13 [2007/08/07] changed "return" to make them consistent
# 0.12 [2007/08/06] changed _load_list and _load_detail to convert charcode
# 0.11 [2007/08/03] changed driver_param, _load_id, and _save_id to use new parameters
# 0.10 [2007/07/31] renamed and modified for Serene Bach 3.*
# 0.09 [2006/11/09] changed _sort_object/_encode to make them faster
# 0.08 [2006/11/07] changed _load_detail and _load_list to make them faster
# 0.07 [2006/02/03] added decrement_id, changed _sort_object to sort data correctly
# 0.06 [2005/11/07] changed _save_list to update correctly
# 0.05 [2005/10/20] changed _generate_condition to check faster
# 0.04 [2005/10/19] changed _load_detail to load data correctly
# 0.03 [2005/10/19] use ref instead of &_class, removed _class
# 0.02 [2005/08/23] changed _load_detail to separate loading an object
# 0.01 [2005/03/01] changed how to lock a file
# 0.00 [2005/02/02] generated
# --------------------------------------
# // defines / static variables
sub DEFAULT_DIRECTORY  (){ './data/' };
sub DEFAULT_SUFFIX     (){ '.cgi' };
sub DEFAULT_INDEX      (){ 'id' };
sub DEFAULT_PERMISSION (){ 0666 };
sub DEFAULT_LOCKFILE   (){ 'save' };
sub DEFAULT_UNIT       (){ 100 };
sub DEFAULT_CODE       (){ 'euc' };
sub TEMPORARY_FILE     (){ '_tmp' };
my %ContentOfList = (
	'entry'     => 'id,wid,subj,cat,date,auth,stat,com,tb,file,tz,add,',
	'message'   => 'id,wid,eid,stat,date,auth,host,tz,',
	'trackback' => 'id,wid,eid,stat,date,subj,name,url,tz,',
	'template'  => 'id,wid,use,name,gen,mod,info,',
	'user'      => 'id,wid,name,pass,real,disp,mail,notice,stat,order,',
);
# --------------------------------------
# // override base class - param
sub driver_param
{
	my $class = shift;
	return (
		'dir'  => DEFAULT_DIRECTORY,
		'suf'  => DEFAULT_SUFFIX,
		'idx'  => DEFAULT_INDEX,
		'mod'  => DEFAULT_PERMISSION,
		'lock' => DEFAULT_LOCKFILE,
		'max'  => {},
		'num'  => 0,
		'code' => DEFAULT_CODE,
	);
}
# --------------------------------------
# // override base class
sub setup
{
	my $self = shift;
	$self->SUPER::setup(
		$self->driver_param(),
		@_,
	);
	my $lock = sb::Lock->basedir . $self->{'lock'};
	if (!-e $lock)
	{
		require sb::File;
		sb::File->new('fullpath'=>$lock)->touch($self->{'mod'});
	}
	$self->{'max'} = $self->_load_id();
	return $self;
}
# --------------------------------------
# // override base class - public method
sub load($$;@)
{
	my $self  = shift;
	my $class = shift; # sets class name of sb::Data::Object
	return if ($class eq '');
	my %param = ( # input paramaeters
		'cond'   => {},     # condition
		'id'     => undef,  # id indicated (return an object with details)
		'logic'  => 'and',  # logic operator for condition
		'bgn'    => 0,      # offset for range
		'num'    => -1,     # number of range
		'sort'   => 'id',   # sorting option
		'order'  => 0,      # order [descent/ascent]
		'detail' => 'on',   # if set true, loading objects with details
		'table'  => ['id'], # elements table of object (this is not an argument)
		@_
	);
	my $func = _generate_condition($param{'logic'},%{$param{'cond'}});
	my @array = $self->_load_detail($class,$func,%param);
	$self->{'num'} = $#array + 1;
	if ( $self->{'num'} > 1 )
	{ # more than one object
		@array = _sort_object($param{'sort'},$param{'order'},@array);
		@array = splice(@array,$param{'bgn'},$param{'num'}) if ( $param{'num'} > 0 and @array > $param{'num'} );
	}
	return( @array );
}
sub count($$;@)
{
	my $self = shift;
	# sb::Driver::TextOld does not have proper count method.
	# So we should call "load" to count matched objects.
	$self->load(@_);
	return $self->matched_number();
}
sub save($@)
{
	my $self = shift;
	my @objs = @_; # set instances of sb::Data::Object to save
	return if ( $#objs == -1 );
	my $lock = sb::Lock->basedir . $self->{'lock'};
	my $class = ref($objs[0]);
	@objs = sort { $a->id <=> $b->id } @objs;
	my $lfh = sb::Lock->locked_open($lock) or die(sb::Language->get->string('Failed to lock'));
	$self->_save_detail($class,@objs);
	$self->_save_list($class,@objs);
	$self->_save_id();
	close($lfh);
	return 1;
}
sub insert($$)
{
	my $self = shift;
	my $obj = shift; # set an instance of sb::Data::Object to save
	my $class = ref($obj);
	if ( !defined($self->{'max'}->{_name($class)})
	  or $obj->id > $self->{'max'}->{_name($class)} )
	{
		$self->{'max'}->{$self->name($class)} = $obj->id;
	}
	$self->save($obj);
	return 1;
}
sub new_id($$)
{
	my $self = shift;
	my $class = shift; # set class name of sb::Data::Object
	return if ($class eq '');
	my $name = _name($class);
	return $self->{'max'}->{$name}++;
}
sub create_table($$)
{
	my $self  = shift;
	my $class = shift; # set class name of sb::Data::Object
	my $file  = $self->{'dir'} . _name($class) . $self->{'suf'};
	require sb::File;
	sb::File->new('fullpath'=>$file)->touch($self->{'mod'});
}
sub alter($$$;$)
{
	my $self = shift;
	my $class = shift; # sets class name of sb::Data::Object
	my $target = shift;
	my $mode = shift; # CHANGE / ADD / DROP
	$mode = 'CHANGE' if ( $mode ne 'ADD' and $mode ne 'DROP' );
	return 1;
}
# --------------------------------------
# // deprecated functions
sub decrement_id
{
	my $self = shift;
	my $class = shift;
	return if ($class eq '');
	my $name = _name($class);
	$self->{'max'}->{$name}--;
	$self->{'max'}->{$name} = 0 if ($self->{'max'}->{$name} < 0);
	return $self->{'max'}->{$name};
}
# --------------------------------------
# // private functions - instance method
sub _save_detail
{
	my $self  = shift;
	my $class = shift;
	my @objs  = @_;
	my $name  = _name($class);
	return if ( !defined($ContentOfList{$name}) );
	my $dir = $self->{'dir'} . $name . '/';
	foreach my $obj ( @objs )
	{
		next if (ref($obj) ne $class);
		my $file = $dir . $obj->id . $self->{'suf'};
		if ($obj->erased)
		{
			unlink($file);
		}
		else
		{
			open(DATAOUT,">$file") or die(sb::Language->get->string('Failed to open the file') . ' : ' . $file);
			binmode(DATAOUT);
			print DATAOUT _encode($obj,'');
			close(DATAOUT);
			chmod($self->{'mod'},$file);
		}
	}
	return;
}
sub _save_list
{
	my $self  = shift;
	my $class = shift;
	my @objs  = @_;
	my $name  = _name($class);
	my $file  = $self->{'dir'} . $name . $self->{'suf'};
	my $temp  = $self->{'dir'} . $name . TEMPORARY_FILE;
	open(LISTOUT,">$temp") or die(sb::Language->get->string('Failed to open the file') . ' : ' . $temp);
	open(LISTIN,"<$file")  or die(sb::Language->get->string('Failed to open the file') . ' : ' . $file);
	binmode(LISTOUT);
	while (my $line = <LISTIN>)
	{
		my @data = _decode($line);
		if (@objs and $data[0] == $objs[0]->id)
		{
			print LISTOUT _encode($objs[0],$name) if (!$objs[0]->erased and ref($objs[0]) eq $class);
			shift(@objs);
		}
		else
		{
			print LISTOUT $line;
		}
	}
	close(LISTIN);
	foreach my $obj ( @objs )
	{
		next if ($obj->erased or ref($obj) ne $class);
		print LISTOUT _encode($obj,$name);
	}
	close(LISTOUT);
	rename($temp,$file);
	chmod($self->{'mod'},$file);
}
sub _load_detail
{
	my $self  = shift;
	my $class = shift;
	my $func  = shift;
	my %param = @_;
	my @array = ();
	my $name  = _name($class);
	return( $self->_load_list($class,$func,%param) ) if ( _switch_loadtype($name,%param) );
	my $lang  = sb::Language->get;
	my $conv  = ($self->{'code'} ne $lang->charcode);
	my $dir = $self->{'dir'} . $name . '/';
	if ( defined($param{'id'}) )
	{
		my $file = $dir . $param{'id'} . $self->{'suf'};
		return if (!-r $file or -z $file);
		open(DATAIN,"<$file");
		my $line = <DATAIN>;
		$line = $lang->convert($line,$lang->charcode,$self->{'code'}) if ($conv);
		my $obj = $class->create_with_array(_decode($line));
		close(DATAIN);
		push(@array,$obj);
	}
	else
	{
		opendir(DATADIR, $dir);
		my @file_list = readdir(DATADIR);
		closedir(DATADIR);
		foreach my $file_name (@file_list)
		{
			next if ($file_name !~ /^(\d+)\./);
			my $file = $dir . $file_name;
			open(DATAIN,"<$file");
			my $line = <DATAIN>;
			$line = $lang->convert($line,$lang->charcode,$self->{'code'}) if ($conv);
			my $obj = $class->create_with_array(_decode($line));
			close(DATAIN);
			next if ( &{$func}($obj) );
			push(@array,$obj);
		}
	}
	return( @array );
}
sub _load_list
{
	my $self  = shift;
	my $class = shift;
	my $func  = shift;
	my %param = @_;
	my @array = ();
	my $name  = _name($class);
	my $file  = $self->{'dir'} . $name . $self->{'suf'};
	my $lang  = sb::Language->get;
	my $conv  = ($self->{'code'} ne $lang->charcode);
	return( @array ) if (!-e $file);
	my @buffer = ();
	open(DATAIN,"<$file");
	while (my $line = <DATAIN>)
	{
		$line = $lang->convert($line,$lang->charcode,$self->{'code'}) if ($conv);
		my @data = _decode($line);
		next if (defined($param{'id'}) and $param{'id'} != $data[0]);
		my $obj = $class->create_with_array(@data);
		next if ( &{$func}($obj) );
		push(@array,$obj);
		last if ( defined($param{'id'}) );
	}
	close(DATAIN);
	return( @array );
}
sub _switch_loadtype
{
	my $name  = shift;
	my %param = @_;
	return 1 if ( !defined($ContentOfList{$name}) );
	return if ( defined($param{'id'}) );
	return if ( defined($param{'detail'}) and $param{'detail'} eq 'on' );
	foreach my $key ( keys(%{$param{'cond'}}) )
	{
		return if ( index($ContentOfList{$name},$key . ',') == -1 );
	}
	return 1;
}
sub _generate_condition
{
	my $logic = shift;
	my %cond  = @_;
	return( sub { return(0) } ) if ( !keys(%cond) );
	my %reg = ();
	foreach my $key ( keys(%cond) )
	{
		next if ( $key eq '__range' or $key eq '__combo' );
		if ( $cond{$key} =~ /^\/(.*)\/$/ )
		{
			$reg{$key} = $1;
			$cond{$key} = undef;
		}
	}
	my $re = ( keys(%reg) ) ? 1 : 0;
	my $func  = sub {
		my $flag = 0;
		foreach my $key ( keys(%reg) )
		{
			my $word = $reg{$key};
			$flag = ( index(lc($_[0]->{$key}),lc($word)) > -1 ); # ( $_[0]->{$key} =~ /\Q$word\E/i )
			last if ($flag);
		}
		return( !$flag ) if ( ($flag and $logic eq 'or' and $re) or (!$flag and $logic eq 'and' and $re) );
		foreach my $key ( keys(%cond) )
		{
			next if ( $key eq '__range' or $key eq '__combo' );
			next if ( !defined($cond{$key}) or !defined($_[0]->{$key}) );
			if ( ref($cond{$key}) eq 'ARRAY' )
			{
				my $check = $_[0]->{$key};
				$check += sb::Time->diff_timezone($_[0]->{'tz'}, 0) if ( $cond{'__range'}->{$key} eq 'tz' );
				$flag = ( defined( $cond{'__range'}->{$key} ) ) 
				      ? ( $check >= $cond{$key}[0] and $check < $cond{$key}[1] )
				      : ( grep( /^\Q$check\E$/, @{$cond{$key}} ) );
			}
			else
			{
				$flag = ( $_[0]->{$key} eq $cond{$key} );
			}
			if ( defined( $cond{'__combo'}->{$key} ) and !$flag )
			{
				my %combo = %{$cond{'__combo'}};
				$flag = ( index($_[0]->{$combo{$key}},$combo{$combo{$key}}) > -1 );
			}
			last if ( ($flag and $logic eq 'or') or (!$flag and $logic eq 'and') );
		}
		return( !$flag );
	};
	return( $func );
}
sub _name
{
	return lc( (split('::',$_[0]))[-1] );
}
sub _decode
{
	my $line = shift;
	$line =~ tr/\x0D\x0A//d;
	return map { s/\\(.)/$1 eq 't' and "\t" or $1 eq 'n' and "\n" or "$1"/eg; $_; } split("\t",$line);
}
sub _encode
{
	my ($obj,$name) = @_;
	my @data = ();
	my @elements = ( $name and defined($ContentOfList{$name}) ) 
	             ? split(',',$ContentOfList{$name}) 
	             : $obj->elements;
	foreach my $elem ( @elements )
	{
		push(@data,_linefeed($obj->{$elem}));
	}
	return join("\t",map { s/\\/\\\\/g; s/\t/\\t/g; s/\n/\\n/g; $_; } @data) . "\t\n";
}
sub _linefeed
{ # unify linefeed code
	$_[0] =~ s/\x0D\x0A/\n/g;
	$_[0] =~ tr/\x0D\x0A/\n\n/;
	while ($_[0] =~ /\n$/) { chomp($_[0]); }
	return $_[0];
}
sub _load_id
{
	my $self = shift;
	my $file = $self->{'dir'} . $self->{'idx'} . $self->{'suf'};
	my %max  = ();
	open(MAXIN,"<$file") or die(sb::Language->get->string('Failed to open the file') . ' : ' . $file);
	while (my $line = <MAXIN>)
	{
		$line =~ tr/\x0D\x0A//d;
		my ($key,$val) = split("\t",$line,2);
		next if ($key eq '');
		$max{$key} = int($val);
	}
	close(MAXIN);
	return \%max;
}
sub _save_id
{
	my $self = shift;
	my %max  = %{$self->{'max'}};
	my $file = $self->{'dir'} . $self->{'idx'} . $self->{'suf'};
	open(MAXOUT,">$file") or die(sb::Language->get->string('Failed to open the file') . ' : ' . $file);
	binmode(MAXOUT);
	foreach my $key ( keys(%max) )
	{
		print MAXOUT $key,"\t",$max{$key},"\n";
	}
	close(MAXOUT);
	return;
}
sub _sort_object
{
	my $elem  = shift;
	my $order = shift;
	my @array = @_;
	my (@data,@ids);
	foreach (@array)
	{
		push(@data, $_->{$elem});
		push(@ids,  $_->{'id'});
	}
	return @array[ sort {
		$data[$b] <=> $data[$a] or $data[$a] cmp $data[$b] or $ids[$b] <=> $ids[$a]
	} 0 .. $#ids ] if ($order);
	return @array[ sort {
		$data[$a] <=> $data[$b] or $data[$b] cmp $data[$a] or $ids[$a] <=> $ids[$b]
	} 0 .. $#ids ];
}
1; # end of package
__END__
